@@ -1,6 +1,6 @@
 /*
    This file is part of GNUnet.
-   Copyright (C) 2021--2022 GNUnet e.V.
+   Copyright (C) 2021--2024 GNUnet e.V.
 
    GNUnet is free software: you can redistribute it and/or modify it
    under the terms of the GNU Affero General Public License as published
@@ -24,19 +24,22 @@
 
 #include "event.h"
 
+#include "application.h"
 #include "contact.h"
 #include "file.h"
 #include "ui.h"
 
-#include "ui/account_entry.h"
 #include "ui/chat_entry.h"
-#include "ui/contact_entry.h"
 #include "ui/message.h"
+#include <gnunet/gnunet_chat_lib.h>
+#include <gnunet/gnunet_common.h>
 
 static void
 _close_notification(NotifyNotification* notification,
-		    gpointer user_data)
+		                gpointer user_data)
 {
+  g_assert((notification) && (user_data));
+
   MESSENGER_Application *app = (MESSENGER_Application*) user_data;
 
   app->notifications = g_list_remove(app->notifications, notification);
@@ -49,19 +52,24 @@ _close_notification(NotifyNotification* notification,
 
 static void
 _show_notification(MESSENGER_Application *app,
-		   UNUSED struct GNUNET_CHAT_Context *context,
-		   const struct GNUNET_CHAT_Contact *contact,
-		   const char *text,
-		   const char *icon,
-		   const char *category)
+                   UNUSED struct GNUNET_CHAT_Context *context,
+                   const struct GNUNET_CHAT_Contact *contact,
+                   const char *text,
+                   const char *icon,
+                   const char *category)
 {
+  g_assert(
+    (app) && (context) && (contact) && 
+    (text) && (icon) && (category)
+  );
+
   if (app->settings.disable_notifications)
     return;
 
   const char *sender = GNUNET_CHAT_contact_get_name(contact);
 
   NotifyNotification *notification = notify_notification_new(
-      sender? sender : "(unknown)", text, icon
+    sender? sender : "(unknown)", text, icon
   );
 
   if (!notification)
@@ -72,10 +80,10 @@ _show_notification(MESSENGER_Application *app,
   notify_notification_set_category(notification, category);
 
   g_signal_connect(
-      notification,
-      "closed",
-      G_CALLBACK(_close_notification),
-      app
+    notification,
+    "closed",
+    G_CALLBACK(_close_notification),
+    app
   );
 
   notify_notification_show(notification, NULL);
@@ -83,30 +91,34 @@ _show_notification(MESSENGER_Application *app,
 
 void
 event_handle_warning(MESSENGER_Application *app,
-		     struct GNUNET_CHAT_Context *context,
-		     const struct GNUNET_CHAT_Message *msg)
+		                 struct GNUNET_CHAT_Context *context,
+		                 const struct GNUNET_CHAT_Message *msg)
 {
+  g_assert((app) && (context) && (msg));
+
   const char *text = GNUNET_CHAT_message_get_text(msg);
 
   const struct GNUNET_CHAT_Contact *contact = GNUNET_CHAT_message_get_sender(
-      msg
+    msg
   );
 
   g_printerr("ERROR: %s\n", text);
 
   _show_notification(
-      app,
-      context,
-      contact,
-      text,
-      "dialog-warning-symbolic",
-      "im.error"
+    app,
+    context,
+    contact,
+    text,
+    "dialog-warning-symbolic",
+    "im.error"
   );
 }
 
 static gboolean
 _idle_refresh_accounts(gpointer user_data)
 {
+  g_assert(user_data);
+
   MESSENGER_Application *app = (MESSENGER_Application*) user_data;
 
   if (!(app->ui.messenger.main_window))
@@ -125,13 +137,15 @@ refresh_exit:
 void
 event_refresh_accounts(MESSENGER_Application *app)
 {
+  g_assert(app);
+
   if (app->ui.messenger.account_refresh)
     g_source_remove(app->ui.messenger.account_refresh);
 
   if (app->ui.messenger.main_window)
     app->ui.messenger.account_refresh = g_idle_add(
-	G_SOURCE_FUNC(_idle_refresh_accounts),
-	app
+      G_SOURCE_FUNC(_idle_refresh_accounts),
+      app
     );
   else
     app->ui.messenger.account_refresh = 0;
@@ -140,6 +154,8 @@ event_refresh_accounts(MESSENGER_Application *app)
 static gboolean
 _select_chat_to_activate(gpointer user_data)
 {
+  g_assert(user_data);
+
   UI_CHAT_ENTRY_Handle *entry = (UI_CHAT_ENTRY_Handle*) user_data;
 
   if ((!entry) || (!(entry->chat)) || (!(entry->entry_box)))
@@ -151,7 +167,7 @@ _select_chat_to_activate(gpointer user_data)
     return FALSE;
 
   GtkListBoxRow *row = GTK_LIST_BOX_ROW(
-      gtk_widget_get_parent(entry->entry_box)
+    gtk_widget_get_parent(entry->entry_box)
   );
 
   if (!row)
@@ -171,6 +187,8 @@ _select_chat_to_activate(gpointer user_data)
 static gboolean
 _idle_chat_entry_update(gpointer user_data)
 {
+  g_assert(user_data);
+
   UI_CHAT_ENTRY_Handle *entry = (UI_CHAT_ENTRY_Handle*) user_data;
 
   if ((!(entry->chat)) || (!(entry->chat->app)) ||
@@ -178,10 +196,10 @@ _idle_chat_entry_update(gpointer user_data)
     goto update_exit;
 
   struct GNUNET_CHAT_Context *context = (struct GNUNET_CHAT_Context*) (
-      g_object_get_qdata(
-	  G_OBJECT(entry->chat->send_text_view),
-	  entry->chat->app->quarks.data
-      )
+    g_object_get_qdata(
+      G_OBJECT(entry->chat->send_text_view),
+      entry->chat->app->quarks.data
+    )
   );
 
   ui_chat_entry_update(entry, entry->chat->app, context);
@@ -194,19 +212,23 @@ update_exit:
 static void
 enqueue_chat_entry_update(UI_CHAT_ENTRY_Handle *entry)
 {
+  g_assert(entry);
+
   if (entry->update)
     g_source_remove(entry->update);
 
   entry->update = g_idle_add(
-      G_SOURCE_FUNC(_idle_chat_entry_update),
-      entry
+    G_SOURCE_FUNC(_idle_chat_entry_update),
+    entry
   );
 }
 
 static void
 _add_new_chat_entry(MESSENGER_Application *app,
-		    struct GNUNET_CHAT_Context *context)
+		                struct GNUNET_CHAT_Context *context)
 {
+  g_assert((app) && (context));
+
   UI_MESSENGER_Handle *ui = &(app->ui.messenger);
   UI_CHAT_ENTRY_Handle *entry = ui_chat_entry_new(app);
 
@@ -216,14 +238,14 @@ _add_new_chat_entry(MESSENGER_Application *app,
   GNUNET_CHAT_context_set_user_pointer(context, entry);
 
   gtk_container_add(
-      GTK_CONTAINER(ui->chats_stack),
-      entry->chat->chat_box
+    GTK_CONTAINER(ui->chats_stack),
+    entry->chat->chat_box
   );
 
   g_object_set_qdata(
-      G_OBJECT(entry->chat->send_text_view),
-      app->quarks.data,
-      context
+    G_OBJECT(entry->chat->send_text_view),
+    app->quarks.data,
+    context
   );
 
   ui->chat_entries = g_list_append(ui->chat_entries, entry);
@@ -231,29 +253,31 @@ _add_new_chat_entry(MESSENGER_Application *app,
   GtkWidget *row = gtk_widget_get_parent(entry->entry_box);
 
   g_object_set_qdata(
-      G_OBJECT(row),
-      app->quarks.ui,
-      entry
+    G_OBJECT(row),
+    app->quarks.ui,
+    entry
   );
 
   if (ui->chat_selection)
     g_source_remove(ui->chat_selection);
 
   ui->chat_selection = g_idle_add(
-      G_SOURCE_FUNC(_select_chat_to_activate),
-      entry
+    G_SOURCE_FUNC(_select_chat_to_activate),
+    entry
   );
 }
 
 static int
 _iterate_profile_contacts(void *cls,
-			  UNUSED struct GNUNET_CHAT_Handle *handle,
-			  struct GNUNET_CHAT_Contact *contact)
+			                    UNUSED struct GNUNET_CHAT_Handle *handle,
+			                    struct GNUNET_CHAT_Contact *contact)
 {
+  g_assert((cls) && (contact));
+
   MESSENGER_Application *app = (MESSENGER_Application*) cls;
 
   struct GNUNET_CHAT_Context *context = GNUNET_CHAT_contact_get_context(
-      contact
+    contact
   );
 
   if (GNUNET_SYSERR == GNUNET_CHAT_context_get_status(context))
@@ -265,13 +289,15 @@ _iterate_profile_contacts(void *cls,
 
 static int
 _iterate_profile_groups(void *cls,
-			UNUSED struct GNUNET_CHAT_Handle *handle,
-			UNUSED struct GNUNET_CHAT_Group *group)
+		                    UNUSED struct GNUNET_CHAT_Handle *handle,
+		                    UNUSED struct GNUNET_CHAT_Group *group)
 {
+  g_assert(cls);
+
   MESSENGER_Application *app = (MESSENGER_Application*) cls;
 
   struct GNUNET_CHAT_Context *context = GNUNET_CHAT_group_get_context(
-      group
+    group
   );
 
   if (GNUNET_SYSERR == GNUNET_CHAT_context_get_status(context))
@@ -283,17 +309,29 @@ _iterate_profile_groups(void *cls,
 
 static void
 _clear_chat_entry(GtkWidget *widget,
-		  gpointer user_data)
+		              gpointer user_data)
 {
+  g_assert((widget) && (user_data));
+
   MESSENGER_Application *app = (MESSENGER_Application*) user_data;
+  UI_MESSENGER_Handle *ui = &(app->ui.messenger);
+
   GtkListBoxRow *row = GTK_LIST_BOX_ROW(widget);
 
   if (!gtk_list_box_row_get_selectable(row))
     return;
 
   UI_CHAT_ENTRY_Handle *entry = (UI_CHAT_ENTRY_Handle*) g_object_get_qdata(
-      G_OBJECT(row),
-      app->quarks.ui
+    G_OBJECT(row),
+    app->quarks.ui
+  );
+
+  ui->chat_entries = g_list_remove(ui->chat_entries, entry);
+
+  g_object_set_qdata(
+    G_OBJECT(entry->chat->send_text_view),
+    app->quarks.data,
+    NULL
   );
 
   ui_chat_entry_dispose(entry, app);
@@ -302,6 +340,8 @@ _clear_chat_entry(GtkWidget *widget,
 void
 event_update_profile(MESSENGER_Application *app)
 {
+  g_assert(app);
+
   UI_MESSENGER_Handle *ui = &(app->ui.messenger);
   CHAT_MESSENGER_Handle *chat = &(app->chat.messenger);
 
@@ -314,21 +354,46 @@ event_update_profile(MESSENGER_Application *app)
 
   ui_label_set_text(ui->profile_key_label, key);
 
-  gtk_container_foreach(
-      GTK_CONTAINER(ui->chats_listbox),
-      _clear_chat_entry,
-      app
+  GList *entries = gtk_container_get_children(
+    GTK_CONTAINER(ui->chats_listbox)
   );
 
+  g_list_foreach(entries, (GFunc) _clear_chat_entry, app);
+  g_list_free(entries);
+
   gtk_stack_set_visible_child(ui->chats_stack, ui->no_chat_box);
 
   GNUNET_CHAT_iterate_contacts(chat->handle, _iterate_profile_contacts, app);
   GNUNET_CHAT_iterate_groups(chat->handle, _iterate_profile_groups, app);
 }
 
+static int
+_cleanup_profile_contacts(void *cls,
+			                    UNUSED struct GNUNET_CHAT_Handle *handle,
+			                    struct GNUNET_CHAT_Contact *contact)
+{
+  g_assert((cls) && (contact));
+
+  if (contact)
+    contact_destroy_info(contact);
+  return GNUNET_YES;
+}
+
+void
+event_cleanup_profile(MESSENGER_Application *app)
+{
+  g_assert(app);
+
+  CHAT_MESSENGER_Handle *chat = &(app->chat.messenger);
+
+  GNUNET_CHAT_iterate_contacts(chat->handle, _cleanup_profile_contacts, NULL);
+}
+
 gboolean
 _delayed_context_drop(gpointer user_data)
 {
+  g_assert(user_data);
+
   struct GNUNET_CHAT_Context *context = (struct GNUNET_CHAT_Context*) user_data;
 
   struct GNUNET_CHAT_Group *group = GNUNET_CHAT_context_get_group(context);
@@ -344,13 +409,15 @@ _delayed_context_drop(gpointer user_data)
 
 void
 event_update_chats(MESSENGER_Application *app,
-		   struct GNUNET_CHAT_Context *context,
-		   const struct GNUNET_CHAT_Message *msg)
+		               struct GNUNET_CHAT_Context *context,
+		               const struct GNUNET_CHAT_Message *msg)
 {
+  g_assert((app) && (context) && (msg));
+
   UI_CHAT_ENTRY_Handle *handle = GNUNET_CHAT_context_get_user_pointer(context);
 
   const enum GNUNET_CHAT_MessageKind kind = GNUNET_CHAT_message_get_kind(
-      msg
+    msg
   );
 
   if (GNUNET_CHAT_KIND_JOIN == kind)
@@ -362,12 +429,12 @@ event_update_chats(MESSENGER_Application *app,
 
     if (app->settings.leave_chats_delay > 0)
       g_timeout_add_seconds(
-	  app->settings.leave_chats_delay,
-	  _delayed_context_drop,
-	  context
+        app->settings.leave_chats_delay,
+        _delayed_context_drop,
+        context
       );
   }
-  else if (handle)
+  else if ((handle) && (handle->entry_box))
     _clear_chat_entry(gtk_widget_get_parent(handle->entry_box), app);
 
   contact_create_info(GNUNET_CHAT_message_get_sender(msg));
@@ -376,8 +443,10 @@ event_update_chats(MESSENGER_Application *app,
 static void
 _update_contact_context(struct GNUNET_CHAT_Contact *contact)
 {
+  g_assert(contact);
+
   struct GNUNET_CHAT_Context *context = GNUNET_CHAT_contact_get_context(
-      contact
+    contact
   );
 
   if (!context)
@@ -393,65 +462,71 @@ _update_contact_context(struct GNUNET_CHAT_Contact *contact)
 
 void
 event_presence_contact(MESSENGER_Application *app,
-		       struct GNUNET_CHAT_Context *context,
-		       const struct GNUNET_CHAT_Message *msg)
+		                   struct GNUNET_CHAT_Context *context,
+		                   const struct GNUNET_CHAT_Message *msg)
 {
+  g_assert((app) && (context) && (msg));
+
   UI_CHAT_ENTRY_Handle *handle = GNUNET_CHAT_context_get_user_pointer(context);
 
-  if (!handle)
-    return;
-
   struct GNUNET_CHAT_Contact *contact = GNUNET_CHAT_message_get_sender(
-      msg
+    msg
   );
 
   if (!contact)
     return;
 
-  const enum GNUNET_CHAT_MessageKind kind = GNUNET_CHAT_message_get_kind(
-      msg
-  );
+  contact_create_info(contact);
+
+  if (!handle)
+    return;
 
   struct GNUNET_CHAT_Group *group = GNUNET_CHAT_context_get_group(context);
 
   UI_MESSAGE_Handle *message = NULL;
-
-  contact_create_info(contact);
+  gboolean previous_presence = FALSE;
 
   if (group)
     message = (UI_MESSAGE_Handle*) (
-	GNUNET_CHAT_member_get_user_pointer(group, contact)
+	    GNUNET_CHAT_member_get_user_pointer(group, contact)
     );
   else
     message = (UI_MESSAGE_Handle*) contact_get_last_message_from_info(contact);
 
   if (message)
+  {
     ui_chat_remove_message(handle->chat, app, message);
+    previous_presence = TRUE;
+  }
 
   message = ui_message_new(app, UI_MESSAGE_STATUS);
   ui_message_update(message, app, msg);
 
   _update_contact_context(contact);
+  ui_message_set_contact(message, contact);
 
-  contact_add_name_avatar_to_info(contact, message->sender_avatar);
-  contact_add_name_label_to_info(contact, message->sender_label);
-
-  const char *text = (
-      GNUNET_CHAT_KIND_JOIN == kind? _("joined the chat") : _("left the chat")
+  const enum GNUNET_CHAT_MessageKind kind = GNUNET_CHAT_message_get_kind(
+    msg
   );
 
-  if (!ui_messenger_is_context_active(&(app->ui.messenger), context))
+  const char *text = (
+    GNUNET_CHAT_KIND_JOIN == kind? _("joined the chat") : _("left the chat")
+  );
+
+  if ((!ui_messenger_is_context_active(&(app->ui.messenger), context)) &&
+      ((!previous_presence) || (GNUNET_CHAT_KIND_LEAVE == kind)) &&
+      (GNUNET_YES == GNUNET_CHAT_message_is_recent(msg)))
     _show_notification(
-	app,
-	context,
-	contact,
-	text,
-	"avatar-default-symbolic",
-	"presence.online"
+      app,
+      context,
+      contact,
+      text,
+      "avatar-default-symbolic",
+      "presence.online"
     );
 
   struct GNUNET_TIME_Absolute timestamp = GNUNET_CHAT_message_get_timestamp(
-      msg
+    msg
   );
 
   const char *time = GNUNET_STRINGS_absolute_time_to_string(timestamp);
@@ -471,11 +546,13 @@ event_presence_contact(MESSENGER_Application *app,
 
 void
 event_update_contacts(UNUSED MESSENGER_Application *app,
-		      struct GNUNET_CHAT_Context *context,
-		      const struct GNUNET_CHAT_Message *msg)
+		                  struct GNUNET_CHAT_Context *context,
+		                  const struct GNUNET_CHAT_Message *msg)
 {
+  g_assert((app) && (context) && (msg));
+
   struct GNUNET_CHAT_Contact *contact = GNUNET_CHAT_message_get_sender(
-      msg
+    msg
   );
 
   if (!contact)
@@ -494,20 +571,37 @@ event_update_contacts(UNUSED MESSENGER_Application *app,
 
 static void
 _event_invitation_accept_click(UNUSED GtkButton *button,
-			       gpointer user_data)
+			                         gpointer user_data)
 {
+  g_assert(user_data);
+
   struct GNUNET_CHAT_Invitation *invitation = (
-      (struct GNUNET_CHAT_Invitation*) user_data
+    (struct GNUNET_CHAT_Invitation*) user_data
   );
 
   GNUNET_CHAT_invitation_accept(invitation);
 }
 
+static void
+_event_invitation_deny_click(UNUSED GtkButton *button,
+                             gpointer user_data)
+{
+  g_assert(user_data);
+
+  struct GNUNET_CHAT_Invitation *invitation = (
+    (struct GNUNET_CHAT_Invitation*) user_data
+  );
+
+  GNUNET_CHAT_invitation_reject(invitation);
+}
+
 void
 event_invitation(MESSENGER_Application *app,
-		 struct GNUNET_CHAT_Context *context,
-		 const struct GNUNET_CHAT_Message *msg)
+                 struct GNUNET_CHAT_Context *context,
+                 const struct GNUNET_CHAT_Message *msg)
 {
+  g_assert((app) && (context) && (msg));
+
   UI_CHAT_ENTRY_Handle *handle = GNUNET_CHAT_context_get_user_pointer(context);
 
   if (!handle)
@@ -521,8 +615,8 @@ event_invitation(MESSENGER_Application *app,
 
   if (app->settings.delete_invitations_delay > 0)
     GNUNET_CHAT_message_delete(msg, GNUNET_TIME_relative_multiply(
-	GNUNET_TIME_relative_get_second_(),
-	app->settings.delete_invitations_delay
+      GNUNET_TIME_relative_get_second_(),
+      app->settings.delete_invitations_delay
     ));
 
   const int sent = GNUNET_CHAT_message_is_sent(msg);
@@ -533,50 +627,65 @@ event_invitation(MESSENGER_Application *app,
   UI_MESSAGE_Handle *message = ui_message_new(app, UI_MESSAGE_STATUS);
   ui_message_update(message, app, msg);
 
-  const struct GNUNET_CHAT_Contact *contact = GNUNET_CHAT_message_get_sender(
+  const struct GNUNET_CHAT_Contact *sender = GNUNET_CHAT_message_get_sender(
     msg
   );
 
-  contact_add_name_avatar_to_info(contact, message->sender_avatar);
-  contact_add_name_label_to_info(contact, message->sender_label);
-
-  const char *invite_message = _("invited you to a chat");
-
-  if (!ui_messenger_is_context_active(&(app->ui.messenger), context))
-    _show_notification(
-	app,
-	context,
-	contact,
-	invite_message,
-	"mail-message-new-symbolic",
-	"im.received"
-    );
-
-  ui_label_set_text(message->text_label, invite_message);
-
-  g_signal_connect(
-      message->accept_button,
-      "clicked",
-      G_CALLBACK(_event_invitation_accept_click),
-      invitation
+  const struct GNUNET_CHAT_Contact *recipient = GNUNET_CHAT_message_get_recipient(
+    msg
   );
 
-  gtk_widget_show(GTK_WIDGET(message->deny_button));
-  gtk_widget_show(GTK_WIDGET(message->accept_button));
+  ui_message_set_contact(message, sender);
+
+  const char *invite_message = _("invited %s to a chat");
+  const char *recipient_name = (
+    (recipient) && 
+    (GNUNET_YES != GNUNET_CHAT_contact_is_owned(recipient))
+  )? GNUNET_CHAT_contact_get_name(recipient) : _("you");
+
+  GString *message_string = g_string_new(NULL);
+  g_string_printf(message_string, invite_message, recipient_name);
+
+  if ((!ui_messenger_is_context_active(&(app->ui.messenger), context)) &&
+      (GNUNET_YES == GNUNET_CHAT_message_is_recent(msg)))
+    _show_notification(
+	    app,
+      context,
+      sender,
+      message_string->str,
+      "mail-message-new-symbolic",
+      "im.received"
+    );
+
+  ui_label_set_text(message->text_label, message_string->str);
+  g_string_free(message_string, TRUE);
+
+  g_signal_connect(
+    message->accept_button,
+    "clicked",
+    G_CALLBACK(_event_invitation_accept_click),
+    invitation
+  );
+
+  g_signal_connect(
+    message->deny_button,
+    "clicked",
+    G_CALLBACK(_event_invitation_deny_click),
+    invitation
+  );
 
   ui_chat_add_message(handle->chat, app, message);
 
-  if (app->settings.accept_all_invitations)
-    gtk_button_clicked(message->accept_button);
-
   enqueue_chat_entry_update(handle);
 }
 
 void
 event_receive_message(MESSENGER_Application *app,
-		      struct GNUNET_CHAT_Context *context,
-		      const struct GNUNET_CHAT_Message *msg)
+                      struct GNUNET_CHAT_Context *context,
+                      const struct GNUNET_CHAT_Message *msg)
 {
+  g_assert((app) && (context) && (msg));
+
   UI_CHAT_ENTRY_Handle *handle = GNUNET_CHAT_context_get_user_pointer(context);
 
   if (!handle)
@@ -586,22 +695,12 @@ event_receive_message(MESSENGER_Application *app,
 
   if ((sent) && (app->settings.auto_delete_delay > 0))
     GNUNET_CHAT_message_delete(msg, GNUNET_TIME_relative_multiply(
-	GNUNET_TIME_relative_get_second_(),
-	app->settings.auto_delete_delay
+      GNUNET_TIME_relative_get_second_(),
+      app->settings.auto_delete_delay
     ));
 
-  const gboolean whispering = (
-      GNUNET_CHAT_KIND_WHISPER == GNUNET_CHAT_message_get_kind(msg)
-  );
-
-  if ((whispering) && (!(app->settings.show_whispering)))
-    return;
-
   const gchar *text = GNUNET_CHAT_message_get_text(msg);
 
-  if (whispering)
-    text = _("whispers...");
-
   if ((text) && (!(*text)))
     goto skip_message;
 
@@ -609,7 +708,7 @@ event_receive_message(MESSENGER_Application *app,
     GNUNET_CHAT_context_send_read_receipt(context, msg);
 
   const UI_MESSAGE_Type type = (
-      GNUNET_YES == sent? UI_MESSAGE_SENT : UI_MESSAGE_DEFAULT
+    GNUNET_YES == sent? UI_MESSAGE_SENT : UI_MESSAGE_DEFAULT
   );
 
   UI_MESSAGE_Handle *message = ui_message_new(app, type);
@@ -623,35 +722,35 @@ event_receive_message(MESSENGER_Application *app,
 
     if (app->settings.delete_files_delay > 0)
       GNUNET_CHAT_message_delete(msg, GNUNET_TIME_relative_multiply(
-	  GNUNET_TIME_relative_get_second_(),
-	  app->settings.delete_files_delay
+        GNUNET_TIME_relative_get_second_(),
+        app->settings.delete_files_delay
       ));
   }
 
   ui_message_update(message, app, msg);
 
   const struct GNUNET_CHAT_Contact *contact = GNUNET_CHAT_message_get_sender(
-      msg
+    msg
   );
 
-  contact_add_name_avatar_to_info(contact, message->sender_avatar);
-  contact_add_name_label_to_info(contact, message->sender_label);
+  ui_message_set_contact(message, contact);
 
   struct GNUNET_TIME_Absolute timestamp = GNUNET_CHAT_message_get_timestamp(
-      msg
+    msg
   );
 
   const char *time = GNUNET_STRINGS_absolute_time_to_string(timestamp);
 
   if ((!ui_messenger_is_context_active(&(app->ui.messenger), context)) &&
+      (GNUNET_YES == GNUNET_CHAT_message_is_recent(msg)) &&
       (GNUNET_YES != sent))
     _show_notification(
-	app,
-	context,
-	contact,
-	text,
-	"mail-unread-symbolic",
-	"im.received"
+      app,
+      context,
+      contact,
+      text,
+      "mail-unread-symbolic",
+      "im.received"
     );
 
   ui_label_set_text(message->text_label, text);
@@ -663,30 +762,97 @@ skip_message:
   enqueue_chat_entry_update(handle);
 }
 
+static void
+_event_update_tag_message_state(const struct GNUNET_CHAT_Message *msg)
+{
+  g_assert((msg) && (GNUNET_CHAT_KIND_TAG == GNUNET_CHAT_message_get_kind(msg)));
+
+  const struct GNUNET_CHAT_Message *target;
+  target = GNUNET_CHAT_message_get_target(msg);
+
+  if (!target)
+    return;
+
+  const struct GNUNET_CHAT_Contact *contact;
+  contact = GNUNET_CHAT_message_get_sender(target);
+
+  if (contact)
+    contact_update_info(contact);
+}
+
 void
 event_delete_message(MESSENGER_Application *app,
-		     struct GNUNET_CHAT_Context *context,
-		     const struct GNUNET_CHAT_Message *msg)
+                     struct GNUNET_CHAT_Context *context,
+                     const struct GNUNET_CHAT_Message *msg)
 {
+  g_assert((app) && (context) && (msg));
+
   UI_CHAT_ENTRY_Handle *handle = GNUNET_CHAT_context_get_user_pointer(context);
 
   if ((!handle) || (!(handle->chat)))
     return;
 
-  GList *messages = handle->chat->messages;
+  GList *rows = gtk_container_get_children(
+    GTK_CONTAINER(handle->chat->messages_listbox)
+  );
 
-  while (messages)
+  for (GList *row = rows; row; row = row->next)
   {
-    UI_MESSAGE_Handle *message = (UI_MESSAGE_Handle*) (messages->data);
+    UI_MESSAGE_Handle *message = (UI_MESSAGE_Handle*) g_object_get_qdata(
+      G_OBJECT(row->data), app->quarks.ui
+    );
 
-    if ((message) && (message->msg == GNUNET_CHAT_message_get_target(msg)))
+    if ((message) && (message->msg == msg))
     {
       ui_chat_remove_message(handle->chat, app, message);
       break;
     }
-
-    messages = messages->next;
   }
 
+  if (rows)
+    g_list_free(rows);
+
+  if (GNUNET_CHAT_KIND_TAG == GNUNET_CHAT_message_get_kind(msg))
+    _event_update_tag_message_state(msg);
+
+  enqueue_chat_entry_update(handle);
+}
+
+void
+event_tag_message(MESSENGER_Application *app,
+                  struct GNUNET_CHAT_Context *context,
+                  const struct GNUNET_CHAT_Message *msg)
+{
+  g_assert((app) && (context) && (msg));
+
+  UI_CHAT_ENTRY_Handle *handle = GNUNET_CHAT_context_get_user_pointer(context);
+
+  const struct GNUNET_CHAT_Message *target = GNUNET_CHAT_message_get_target(msg);
+
+  _event_update_tag_message_state(msg);
+
+  if ((!handle) || (!(handle->chat)))
+    return;
+
+  GList *rows = gtk_container_get_children(
+    GTK_CONTAINER(handle->chat->messages_listbox)
+  );
+
+  for (GList *row = rows; row; row = row->next)
+  {
+    UI_MESSAGE_Handle *message = (UI_MESSAGE_Handle*) g_object_get_qdata(
+      G_OBJECT(row->data), app->quarks.ui
+    );
+
+    if ((message) && (message->msg == target))
+    {
+      ui_message_update(message, app, message->msg);
+      break;
+    }
+  }
+
+  if (rows)
+    g_list_free(rows);
+
   enqueue_chat_entry_update(handle);
 }
