@@ -1,6 +1,6 @@
 /*
    This file is part of GNUnet.
-   Copyright (C) 2021--2022 GNUnet e.V.
+   Copyright (C) 2021--2024 GNUnet e.V.
 
    GNUnet is free software: you can redistribute it and/or modify it
    under the terms of the GNU Affero General Public License as published
@@ -25,17 +25,21 @@
 #include "message.h"
 
 #include <gnunet/gnunet_chat_lib.h>
+#include <gnunet/gnunet_common.h>
 
 #include "../application.h"
+#include "../contact.h"
 #include "../file.h"
 #include "../ui.h"
 
 static void
 handle_downloading_file(void *cls,
-			const struct GNUNET_CHAT_File *file,
-			uint64_t completed,
-			uint64_t size)
+                        const struct GNUNET_CHAT_File *file,
+                        uint64_t completed,
+                        uint64_t size)
 {
+  g_assert((cls) && (file));
+
   MESSENGER_Application *app = (MESSENGER_Application*) cls;
 
   if (!app)
@@ -46,19 +50,21 @@ handle_downloading_file(void *cls,
 
 static void
 handle_file_button_click(GtkButton *button,
-			 gpointer user_data)
+			                   gpointer user_data)
 {
+  g_assert((button) && (user_data));
+
   MESSENGER_Application *app = (MESSENGER_Application*) user_data;
 
   UI_MESSAGE_Handle* handle = (UI_MESSAGE_Handle*) (
-      g_object_get_qdata(G_OBJECT(button), app->quarks.ui)
+    g_object_get_qdata(G_OBJECT(button), app->quarks.ui)
   );
 
   if (!handle)
     return;
 
   struct GNUNET_CHAT_File *file = (struct GNUNET_CHAT_File*) (
-      g_object_get_qdata(G_OBJECT(handle->file_progress_bar), app->quarks.data)
+    g_object_get_qdata(G_OBJECT(handle->file_progress_bar), app->quarks.data)
   );
 
   if (!file)
@@ -76,9 +82,9 @@ handle_file_button_click(GtkButton *button,
     GNUNET_CHAT_file_stop_download(file);
 
     gtk_image_set_from_icon_name(
-	handle->file_status_image,
-	"folder-download-symbolic",
-	GTK_ICON_SIZE_BUTTON
+      handle->file_status_image,
+      "folder-download-symbolic",
+      GTK_ICON_SIZE_BUTTON
     );
   }
   else if (local_size < size)
@@ -108,9 +114,53 @@ handle_file_button_click(GtkButton *button,
   }
 }
 
+static void
+handle_media_button_click(GtkButton *button,
+			                    gpointer user_data)
+{
+  g_assert((button) && (user_data));
+
+  MESSENGER_Application *app = (MESSENGER_Application*) user_data;
+
+  UI_MESSAGE_Handle* handle = (UI_MESSAGE_Handle*) (
+      g_object_get_qdata(G_OBJECT(button), app->quarks.ui)
+  );
+
+  if (!handle)
+    return;
+
+  struct GNUNET_CHAT_File *file = (struct GNUNET_CHAT_File*) (
+      g_object_get_qdata(G_OBJECT(handle->media_progress_bar), app->quarks.data)
+  );
+
+  if (!file)
+    return;
+
+  const gchar *preview = GNUNET_CHAT_file_open_preview(file);
+
+  if (!preview)
+    return;
+
+  ui_play_media_window_init(app, &(app->ui.play_media));
+
+  GString* uri = g_string_new("file://");
+  g_string_append(uri, preview);
+
+  ui_play_media_window_update(
+      &(app->ui.play_media),
+      uri->str,
+      file
+  );
+
+  gtk_widget_show(GTK_WIDGET(app->ui.play_media.window));
+  g_string_free(uri, TRUE);
+}
+
 static int
 handle_message_redraw_animation(gpointer user_data)
 {
+  g_assert(user_data);
+
   UI_MESSAGE_Handle *handle = (UI_MESSAGE_Handle*) user_data;
 
   handle->redraw_animation = 0;
@@ -126,9 +176,11 @@ handle_message_redraw_animation(gpointer user_data)
 
 static gboolean
 handle_preview_drawing_area_draw(GtkWidget* drawing_area,
-			      cairo_t* cairo,
-			      gpointer user_data)
+                                 cairo_t* cairo,
+                                 gpointer user_data)
 {
+  g_assert((drawing_area) && (cairo) && (user_data));
+
   UI_MESSAGE_Handle *handle = (UI_MESSAGE_Handle*) user_data;
 
   GtkStyleContext* context = gtk_widget_get_style_context(drawing_area);
@@ -147,17 +199,17 @@ handle_preview_drawing_area_draw(GtkWidget* drawing_area,
     gdk_pixbuf_animation_iter_advance(handle->preview_animation_iter, NULL);
   else
     handle->preview_animation_iter = gdk_pixbuf_animation_get_iter(
-	handle->preview_animation, NULL
+	    handle->preview_animation, NULL
     );
 
   image = gdk_pixbuf_animation_iter_get_pixbuf(handle->preview_animation_iter);
 
   const int delay = gdk_pixbuf_animation_iter_get_delay_time(
-      handle->preview_animation_iter
+    handle->preview_animation_iter
   );
 
   handle->redraw_animation = g_timeout_add(
-      delay, handle_message_redraw_animation, handle
+    delay, handle_message_redraw_animation, handle
   );
 
 render_image:
@@ -170,9 +222,9 @@ render_image:
   gint optimal_height = width * dheight / dwidth;
 
   gtk_widget_set_size_request(
-      GTK_WIDGET(drawing_area),
-      width,
-      optimal_height
+    GTK_WIDGET(drawing_area),
+    width,
+    optimal_height
   );
 
   double ratio_width = 1.0 * width / dwidth;
@@ -187,15 +239,15 @@ render_image:
   double dy = (height - dheight) * 0.5;
 
   const int interp_type = (ratio >= 1.0?
-      GDK_INTERP_NEAREST :
-      GDK_INTERP_BILINEAR
+    GDK_INTERP_NEAREST :
+    GDK_INTERP_BILINEAR
   );
 
   GdkPixbuf* scaled = gdk_pixbuf_scale_simple(
-      image,
-      dwidth,
-      dheight,
-      interp_type
+    image,
+    dwidth,
+    dheight,
+    interp_type
   );
 
   gtk_render_icon(context, cairo, scaled, dx, dy);
@@ -209,6 +261,8 @@ render_image:
 static void
 _clear_message_preview_data(UI_MESSAGE_Handle *handle)
 {
+  g_assert(handle);
+
   if (handle->preview_image)
   {
     g_object_unref(handle->preview_image);
@@ -235,22 +289,25 @@ _clear_message_preview_data(UI_MESSAGE_Handle *handle)
 
   if (handle->preview_drawing_area)
     gtk_widget_set_size_request(
-	GTK_WIDGET(handle->preview_drawing_area),
-	-1,
-	-1
+      GTK_WIDGET(handle->preview_drawing_area),
+      -1,
+      -1
     );
 }
 
 UI_MESSAGE_Handle*
 ui_message_new(MESSENGER_Application *app,
-	       UI_MESSAGE_Type type)
+	             UI_MESSAGE_Type type)
 {
+  g_assert(app);
+
   UI_MESSAGE_Handle* handle = g_malloc(sizeof(UI_MESSAGE_Handle));
 
   handle->type = type;
 
   handle->timestamp = GNUNET_TIME_absolute_get_zero_();
   handle->msg = NULL;
+  handle->contact = NULL;
 
   const char *ui_builder_file;
 
@@ -268,25 +325,29 @@ ui_message_new(MESSENGER_Application *app,
   }
 
   handle->builder[0] = gtk_builder_new_from_resource(
-      application_get_resource_path(app, ui_builder_file)
+    application_get_resource_path(app, ui_builder_file)
   );
 
   handle->message_box = GTK_WIDGET(
-      gtk_builder_get_object(handle->builder[0], "message_box")
+    gtk_builder_get_object(handle->builder[0], "message_box")
   );
 
   handle->sender_avatar = HDY_AVATAR(
-      gtk_builder_get_object(handle->builder[0], "sender_avatar")
+    gtk_builder_get_object(handle->builder[0], "sender_avatar")
   );
 
   handle->sender_label = GTK_LABEL(
-      gtk_builder_get_object(handle->builder[0], "sender_label")
+    gtk_builder_get_object(handle->builder[0], "sender_label")
+  );
+
+  handle->private_image = GTK_IMAGE(
+    gtk_builder_get_object(handle->builder[0], "private_image")
   );
 
   if (UI_MESSAGE_STATUS == handle->type)
   {
     handle->deny_revealer = GTK_REVEALER(
-	gtk_builder_get_object(handle->builder[0], "deny_revealer")
+	    gtk_builder_get_object(handle->builder[0], "deny_revealer")
     );
 
     handle->accept_revealer = GTK_REVEALER(
@@ -294,11 +355,11 @@ ui_message_new(MESSENGER_Application *app,
     );
 
     handle->deny_button = GTK_BUTTON(
-	gtk_builder_get_object(handle->builder[0], "deny_button")
+	    gtk_builder_get_object(handle->builder[0], "deny_button")
     );
 
     handle->accept_button = GTK_BUTTON(
-	gtk_builder_get_object(handle->builder[0], "accept_button")
+	    gtk_builder_get_object(handle->builder[0], "accept_button")
     );
   }
   else
@@ -311,73 +372,98 @@ ui_message_new(MESSENGER_Application *app,
   }
 
   GtkContainer *content_box = GTK_CONTAINER(
-      gtk_builder_get_object(handle->builder[0], "content_box")
+    gtk_builder_get_object(handle->builder[0], "content_box")
   );
 
   handle->builder[1] = gtk_builder_new_from_resource(
-      application_get_resource_path(app, "ui/message_content.ui")
+    application_get_resource_path(app, "ui/message_content.ui")
   );
 
   handle->timestamp_label = GTK_LABEL(
-      gtk_builder_get_object(handle->builder[1], "timestamp_label")
+    gtk_builder_get_object(handle->builder[1], "timestamp_label")
   );
 
   handle->read_receipt_image = GTK_IMAGE(
-      gtk_builder_get_object(handle->builder[1], "read_receipt_image")
+    gtk_builder_get_object(handle->builder[1], "read_receipt_image")
   );
 
   handle->content_stack = GTK_STACK(
-      gtk_builder_get_object(handle->builder[1], "content_stack")
+    gtk_builder_get_object(handle->builder[1], "content_stack")
   );
 
   handle->text_label = GTK_LABEL(
-      gtk_builder_get_object(handle->builder[1], "text_label")
+    gtk_builder_get_object(handle->builder[1], "text_label")
   );
 
   handle->file_revealer = GTK_REVEALER(
-      gtk_builder_get_object(handle->builder[1], "file_revealer")
+    gtk_builder_get_object(handle->builder[1], "file_revealer")
   );
 
   handle->filename_label = GTK_LABEL(
-      gtk_builder_get_object(handle->builder[1], "filename_label")
+    gtk_builder_get_object(handle->builder[1], "filename_label")
   );
 
   handle->file_progress_bar = GTK_PROGRESS_BAR(
-      gtk_builder_get_object(handle->builder[1], "file_progress_bar")
+    gtk_builder_get_object(handle->builder[1], "file_progress_bar")
   );
 
   handle->file_button = GTK_BUTTON(
-      gtk_builder_get_object(handle->builder[1], "file_button")
+    gtk_builder_get_object(handle->builder[1], "file_button")
   );
 
   g_signal_connect(
-      handle->file_button,
-      "clicked",
-      G_CALLBACK(handle_file_button_click),
-      app
+    handle->file_button,
+    "clicked",
+    G_CALLBACK(handle_file_button_click),
+    app
   );
 
   handle->file_status_image = GTK_IMAGE(
-      gtk_builder_get_object(handle->builder[1], "file_status_image")
+    gtk_builder_get_object(handle->builder[1], "file_status_image")
   );
 
   g_object_set_qdata(G_OBJECT(handle->file_button), app->quarks.ui, handle);
 
   handle->preview_drawing_area = GTK_DRAWING_AREA(
-      gtk_builder_get_object(handle->builder[1], "preview_drawing_area")
+    gtk_builder_get_object(handle->builder[1], "preview_drawing_area")
   );
 
   g_signal_connect(
-      handle->preview_drawing_area,
-      "draw",
-      G_CALLBACK(handle_preview_drawing_area_draw),
-      handle
+    handle->preview_drawing_area,
+    "draw",
+    G_CALLBACK(handle_preview_drawing_area_draw),
+    handle
   );
 
-  handle->whisper_box = GTK_WIDGET(
-      gtk_builder_get_object(handle->builder[1], "whisper_box")
+  handle->media_revealer = GTK_REVEALER(
+    gtk_builder_get_object(handle->builder[1], "media_revealer")
   );
 
+  handle->media_type_image = GTK_IMAGE(
+    gtk_builder_get_object(handle->builder[1], "media_type_image")
+  );
+
+  handle->media_label = GTK_LABEL(
+    gtk_builder_get_object(handle->builder[1], "media_label")
+  );
+
+  handle->media_progress_bar = GTK_PROGRESS_BAR(
+    gtk_builder_get_object(handle->builder[1], "media_progress_bar")
+  );
+
+  handle->media_button = GTK_BUTTON(
+    gtk_builder_get_object(handle->builder[1], "media_button")
+  );
+
+  g_signal_connect(
+    handle->media_button,
+    "clicked",
+    G_CALLBACK(handle_media_button_click),
+    app
+  );
+
+  g_object_set_qdata(G_OBJECT(handle->media_button), app->quarks.ui, handle);
+
   switch (handle->type)
   {
     case UI_MESSAGE_STATUS:
@@ -388,7 +474,7 @@ ui_message_new(MESSENGER_Application *app,
   }
 
   gtk_container_add(content_box, GTK_WIDGET(
-      gtk_builder_get_object(handle->builder[1], "message_content_box")
+    gtk_builder_get_object(handle->builder[1], "message_content_box")
   ));
 
   handle->preview_image = NULL;
@@ -402,10 +488,12 @@ ui_message_new(MESSENGER_Application *app,
 
 static int
 _iterate_read_receipts(void *cls,
-		       UNUSED const struct GNUNET_CHAT_Message *message,
-		       const struct GNUNET_CHAT_Contact *contact,
-		       int read_receipt)
+                       UNUSED const struct GNUNET_CHAT_Message *message,
+                       const struct GNUNET_CHAT_Contact *contact,
+                       int read_receipt)
 {
+  g_assert((cls) && (message) && (contact));
+
   int *count_read_receipts = (int*) cls;
 
   if ((GNUNET_YES == read_receipt) &&
@@ -418,10 +506,15 @@ _iterate_read_receipts(void *cls,
 void
 ui_message_refresh(UI_MESSAGE_Handle *handle)
 {
+  g_assert(handle);
+
   if ((!(handle->msg)) ||
       (GNUNET_YES != GNUNET_CHAT_message_is_sent(handle->msg)))
     return;
 
+  if (!(handle->read_receipt_image))
+    return;
+
   int count = 0;
   if ((0 < GNUNET_CHAT_message_get_read_receipt(handle->msg, _iterate_read_receipts, &count)) &&
       (0 < count))
@@ -430,11 +523,74 @@ ui_message_refresh(UI_MESSAGE_Handle *handle)
     gtk_widget_hide(GTK_WIDGET(handle->read_receipt_image));
 }
 
+gboolean
+_message_media_supports_file_extension(const gchar *filename)
+{
+  if (!filename)
+    return FALSE;
+
+  const char* extension = strrchr(filename, '.');
+
+  if (!extension)
+    return FALSE;
+
+  if (0 == g_strcmp0(extension, ".ogg"))
+    return TRUE;
+  if (0 == g_strcmp0(extension, ".mp3"))
+    return TRUE;
+  if (0 == g_strcmp0(extension, ".wav"))
+    return TRUE;
+
+  return FALSE;
+}
+
+static void
+_update_invitation_message(UI_MESSAGE_Handle *handle,
+		                       MESSENGER_Application *app,
+                           struct GNUNET_CHAT_Invitation *invitation)
+{
+  g_assert((handle) && (app) && (invitation));
+
+  enum GNUNET_GenericReturnValue accepted, rejected;
+  accepted = GNUNET_CHAT_invitation_is_accepted(invitation);
+  rejected = GNUNET_CHAT_invitation_is_rejected(invitation);
+
+  if (handle->deny_button)
+    gtk_widget_set_sensitive(
+      GTK_WIDGET(handle->deny_button), 
+      GNUNET_YES == accepted || GNUNET_YES == rejected? FALSE : TRUE
+    );
+
+  if (handle->accept_button)
+    gtk_widget_set_sensitive(
+      GTK_WIDGET(handle->accept_button), 
+      GNUNET_YES == accepted || GNUNET_YES == rejected? FALSE : TRUE
+    );
+
+  if (handle->deny_revealer)
+    gtk_revealer_set_reveal_child(
+      GTK_REVEALER(handle->deny_revealer),
+      GNUNET_YES == accepted || GNUNET_YES == rejected? FALSE : TRUE
+    );
+
+  if (handle->accept_revealer)
+    gtk_revealer_set_reveal_child(
+      GTK_REVEALER(handle->accept_revealer),
+      GNUNET_YES == accepted || GNUNET_YES == rejected? FALSE : TRUE
+    );
+
+  if ((app->settings.accept_all_invitations) &&
+      (GNUNET_NO == accepted) && (handle->accept_button))
+    gtk_button_clicked(handle->accept_button);
+}
+
 static void
 _update_file_message(UI_MESSAGE_Handle *handle,
-		     MESSENGER_Application *app,
-		     struct GNUNET_CHAT_File *file)
+		                 MESSENGER_Application *app,
+		                 struct GNUNET_CHAT_File *file)
 {
+  g_assert((handle) && (app) && (file));
+
   const char *filename = GNUNET_CHAT_file_get_name(file);
 
   uint64_t size = GNUNET_CHAT_file_get_size(file);
@@ -451,7 +607,7 @@ _update_file_message(UI_MESSAGE_Handle *handle,
     );
 
     if ((app->settings.accept_all_files) &&
-      (!GNUNET_CHAT_file_is_downloading(file)))
+        (!GNUNET_CHAT_file_is_downloading(file)))
       autostart_download = TRUE;
 
     goto file_content;
@@ -472,7 +628,7 @@ _update_file_message(UI_MESSAGE_Handle *handle,
   if (!(handle->preview_animation))
     handle->preview_image = gdk_pixbuf_new_from_file(preview, NULL);
 
-  if ((handle->preview_animation) || (handle->preview_animation))
+  if ((handle->preview_animation) || (handle->preview_image))
   {
     gtk_widget_set_size_request(
       GTK_WIDGET(handle->preview_drawing_area),
@@ -491,29 +647,61 @@ _update_file_message(UI_MESSAGE_Handle *handle,
 
   GNUNET_CHAT_file_close_preview(file);
 
+  if (_message_media_supports_file_extension(filename))
+  {
+    gtk_image_set_from_icon_name(
+      handle->media_type_image,
+      "audio-x-generic-symbolic",
+      GTK_ICON_SIZE_DND
+    );
+
+    goto media_content;
+  }
+
+  if (!ui_play_media_window_supports_file_extension(filename))
+    goto file_progress;
+
+media_content:
+  ui_label_set_text(handle->media_label, filename);
+
+  gtk_stack_set_visible_child(
+    handle->content_stack,
+    GTK_WIDGET(handle->media_revealer)
+  );
+
+  gtk_revealer_set_reveal_child(handle->media_revealer, TRUE);
+
+  g_object_set_qdata(
+    G_OBJECT(handle->media_progress_bar),
+    app->quarks.data,
+    file
+  );
+
+  return;
+
 file_progress:
   gtk_progress_bar_set_fraction(handle->file_progress_bar, 1.0);
 
   gtk_image_set_from_icon_name(
-      handle->file_status_image,
-      "document-open-symbolic",
-      GTK_ICON_SIZE_BUTTON
+    handle->file_status_image,
+    "document-open-symbolic",
+    GTK_ICON_SIZE_BUTTON
   );
 
 file_content:
   ui_label_set_text(handle->filename_label, filename);
 
   gtk_stack_set_visible_child(
-      handle->content_stack,
-      GTK_WIDGET(handle->file_revealer)
+    handle->content_stack,
+    GTK_WIDGET(handle->file_revealer)
   );
 
   gtk_revealer_set_reveal_child(handle->file_revealer, TRUE);
 
   g_object_set_qdata(
-      G_OBJECT(handle->file_progress_bar),
-      app->quarks.data,
-      file
+    G_OBJECT(handle->file_progress_bar),
+    app->quarks.data,
+    file
   );
 
   if (autostart_download)
@@ -522,10 +710,13 @@ file_content:
 
 void
 ui_message_update(UI_MESSAGE_Handle *handle,
-		  MESSENGER_Application *app,
-		  const struct GNUNET_CHAT_Message *msg)
+		              MESSENGER_Application *app,
+		              const struct GNUNET_CHAT_Message *msg)
 {
+  g_assert((handle) && (app) && (msg));
+
   struct GNUNET_CHAT_File *file = NULL;
+  struct GNUNET_CHAT_Invitation *invitation = NULL;
 
   handle->msg = msg;
 
@@ -533,32 +724,68 @@ ui_message_update(UI_MESSAGE_Handle *handle,
 
   if (msg)
   {
-    if (GNUNET_CHAT_KIND_WHISPER == GNUNET_CHAT_message_get_kind(msg))
-      gtk_stack_set_visible_child(
-	  handle->content_stack,
-	  GTK_WIDGET(handle->whisper_box)
-      );
-
+    if (GNUNET_YES == GNUNET_CHAT_message_is_private(msg))
+      gtk_widget_show(GTK_WIDGET(handle->private_image));
+    
+    invitation = GNUNET_CHAT_message_get_invitation(msg);
     file = GNUNET_CHAT_message_get_file(msg);
 
     handle->timestamp = GNUNET_CHAT_message_get_timestamp(msg);
 
+    if (handle->accept_button)
+      g_object_set_qdata(G_OBJECT(handle->accept_button), app->quarks.data, invitation);
+
     g_object_set_qdata(G_OBJECT(handle->message_box), app->quarks.data, file);
   }
   else
+  {
+    if (handle->accept_button)
+      invitation = (struct GNUNET_CHAT_Invitation*) (
+        g_object_get_qdata(G_OBJECT(handle->accept_button), app->quarks.data)
+      );
+
     file = (struct GNUNET_CHAT_File*) (
-	g_object_get_qdata(G_OBJECT(handle->message_box), app->quarks.data)
+	    g_object_get_qdata(G_OBJECT(handle->message_box), app->quarks.data)
     );
+  }
 
-  if (!file)
-    return;
+  if (invitation)
+    _update_invitation_message(handle, app, invitation);
 
-  _update_file_message(handle, app, file);
+  if (file)
+    _update_file_message(handle, app, file);
+}
+
+void
+ui_message_set_contact(UI_MESSAGE_Handle *handle,
+		                   const struct GNUNET_CHAT_Contact *contact)
+{
+  g_assert(handle);
+
+  if (handle->contact)
+  {
+    contact_remove_name_avatar_from_info(handle->contact, handle->sender_avatar);
+    contact_remove_name_label_from_info(handle->contact, handle->sender_label);
+    contact_remove_visible_widget_to_info(handle->contact, handle->message_box);
+  }
+
+  if (contact)
+  {
+    contact_add_name_avatar_to_info(contact, handle->sender_avatar);
+    contact_add_name_label_to_info(contact, handle->sender_label);
+    contact_add_visible_widget_to_info(contact, handle->message_box);
+  }
+
+  handle->contact = contact;
 }
 
 void
 ui_message_delete(UI_MESSAGE_Handle *handle)
 {
+  g_assert(handle);
+
+  ui_message_set_contact(handle, NULL);
+
   _clear_message_preview_data(handle);
 
   g_object_unref(handle->builder[1]);
